import asyncio
import random

from .exceptions import RemoteMethodError

class MethodCall:
    '''An internal representation of a method call'''
    
    def __init__(self):
        self._id = str(random.randint(1, 1000000))
        self._returned = asyncio.Event()
        
        self._error, self._result = None, None
    
    async def __wait__(self):
        await self._returned.wait()
        
        if self._error:
            raise RemoteMethodError(self._error.get('message', self._error))
        else:
            return self._result

    async def __result__(self, error, result):
        self._error, self._result = error, result
        self._returned.set()



'''Dotable class for using period attribute accessor syntax on Python dicts

Created by Andy Hayden: https://hayd.github.io/2013/dotable-dictionaries
'''

class Dotable(dict):

    __getattr__= dict.__getitem__

    def __init__(self, d):
        self.update(**dict((k, self.parse(v))
                           for k, v in d.items()))

    @classmethod
    def parse(cls, v):
        if isinstance(v, dict):
            return cls(v)
        elif isinstance(v, list):
            return [cls.parse(i) for i in v]
        else:
            return v



import asyncio
import websockets
import random
import ejson

from .subscription import Subscription
from .collection import Collection
from .methodcall import MethodCall
from .exceptions import ConnectionError
        
class DDPClient:
    '''Manages a connection to a server.
    It takes a URL as the first parameter, the optional second parameter specifies which event loop will be used.
    '''
    
    def __init__(self, url, event_loop = None):
        self.connected = False
        self.url = url
        
        self._websocket = None
        self._session = None
        self._event_loop = event_loop or asyncio.get_event_loop()
        
        self._subs = {}
        self._cols = {}
        self._calls = {}
        
    async def connect(self):
        '''This coroutine establishes a connection to a server. It blocks until the connection is established or an exception is raised.'''

        self._websocket = await websockets.connect(self.url)
        
        msg = {'msg': 'connect', 'version': '1', 'support': ['1']}
        if self._session:
            # If we were previously connected, we can attempt to resume that session
            msg['session'] = self._session

        await self._websocket.send(ejson.dumps(msg))

        while self._websocket.open:
            msg = await self._websocket.recv()
            msg = ejson.loads(msg)
            _type = msg.get('msg')
            
            if _type == 'failed':
                raise ConnectionError('The server is not compatible with version 1 of the DDP protocol.')
            elif _type == 'connected':
                self.connected = True
                self._session = msg['session']
            
                self._event_loop.create_task(self.__handler__())

                return
    
    async def disconnect(self):
        await self._websocket.close()
        self.connected = False
        
    async def subscribe(self, name, *params):
        '''Subscribe to a publication.
        subscribe() returns a Subscription object which can be used to monitor the status of the subscription.
        '''
        
        sub = Subscription(name)
        self._subs[sub._id] = sub

        await self._websocket.send(ejson.dumps({
            'msg': 'sub',
            'id': sub._id,
            'name': name,
            'params': params
        }))
        
        return sub
    
    async def unsubscribe(self, sub):
        '''Unsubscribe from a publication.'''
        
        await self._websocket.send(ejson.dumps({
            'msg': 'unsub',
            'id': sub._id
        }))
    
    def get_collection(self, name):
        '''Retrieve an existing Collection by name. If the Collection does not exist it will be created.'''
        c = self._cols.get(name)

        if not c:
            c = Collection(name)
            self._cols[name] = c

        return c
    
    async def call(self, method, *params):
        '''This coroutine calls a remote method on the server and returns the result.
        Raises a ddp_asyncio.RemoteMethodError if the server replies with an error.
        '''
        
        c = MethodCall()
        self._calls[c._id] = c

        await self._websocket.send(ejson.dumps({
            'msg': 'method',
            'method': method,
            'params': params,
            'id': c._id
        }))
        
        return await c.__wait__()
    
    async def __pong__(self, _id):
        '''Respond to a ping from the server.'''
        await self._websocket.send(ejson.dumps({
            'msg': 'pong',
            'id': _id
        }))
                
    async def __handler__(self):
        '''Handles messages received from the server'''

        while self._websocket.open:
            try:
                msg = await self._websocket.recv()
            except websockets.exceptions.ConnectionClosed:
                break
            
            if not msg: continue

            msg = ejson.loads(msg)
            _type = msg.get('msg')
            
            if _type == 'ping':
                await self.__pong__(msg.get('id'))
                
            elif _type == 'ready':
                for _id in msg['subs']:
                    sub = self._subs.get(_id)
                    if sub: sub.__ready__()
            
            elif _type == 'nosub':
                sub = self._subs.get(msg['id'])
                if sub: sub.__error__(msg.get('error', 'Denied by server for unspecified reason.'))

            elif _type == 'added':
                col = self.get_collection(msg['collection'])
                col.__added__(msg['id'], msg['fields'])
                    
            elif _type == 'changed':
                col = self.get_collection(msg['collection'])
                col.__changed__(msg['id'], msg.get('fields', {}), msg.get('cleared', []))
                    
            elif _type == 'removed':
                col = self.get_collection(msg['collection'])
                col.__removed__(msg['id'])
                    
            elif _type == 'result':
                c = self._calls.get(msg['id'])
                if c:
                    await c.__result__(msg.get('error'), msg.get('result'))
                    del self._calls[msg['id']]



import collections
import asyncio
import random
import weakref

from .dotable import Dotable

class CollectionEvent(Dotable):
    '''Represents a change to a Collection.
    
    There are three types of changes: additions, changes, and removals. The type property states the type of change that occurred. 
    
    Additions have the following properties:
        type: 'added'
        _id: id of added item
        fields: contents of added item
    The fields property can be accessed like a dictionary or using dot-notation attribute access.
    
    Changes have the following properties:
        type: 'changed'
        _id: id of changed item
        fields: contents of changed item
        cleared: list of keys removed from the item
    
    Removals have the following properties:
        type: 'removed'
        _id: id of removed item
    '''
    pass

# Note: Collection does not currently support ordered collections, since Meteor doesn't use them.

class Collection(collections.Mapping):
    '''Stores data published from a connected server.

    Collections are read-only and function identically to a dictionary, with each item's _id as the key.

    A Collection item's attributes can be accessed like a dictionarie or with dot-notation attribute access, i.e. "item.attr"
    '''

    def __init__(self, name):
        self._name = name
        self._data = {}

        self._queues = set()

    def __getitem__(self, key):
        return self._data[key]
    
    def __len__(self):
        return len(self._data)
    
    def __iter__(self):
        return iter(self._data)
    
    def __bool__(self):
        return True
    
    def __repr__(self):
        return '<collection {}>'.format(self._name)

    def get_queue(self):
        '''Creates and returns a new asyncio.Queue to monitor changes to a Collection.
        When the collection changes, a CollectionEvent object containing a description of the change is pushed to the queue.
        '''
        
        q = asyncio.Queue()
        self._queues.add(weakref.ref(q))
        return q

    def __put__(self, data):
        for ref in self._queues:
            q = ref()
            if q: q.put_nowait(data)

    def __added__(self, _id, fields):
        self._data[_id] = Dotable.parse(fields)
        self._data[_id]['_id'] = _id
        
        self.__put__(CollectionEvent({
            'type': 'added',
            '_id': _id,
            'fields': Dotable.parse(fields)
        }))
    
    def __changed__(self, _id, fields, cleared):
        if fields:
            self._data[_id].update(fields)
        elif cleared:
            for key in cleared:
                del self._data[_id][key]
        
        self.__put__(CollectionEvent({
            'type': 'changed',
            '_id': _id,
            'fields': Dotable.parse(fields),
            'cleared': cleared
        }))
    
    def __removed__(self, _id):
        del self._data[_id]
        
        self.__put__(CollectionEvent({
            'type': 'removed',
            '_id': _id
        }))



import asyncio
import random

from .exceptions import SubscriptionError

class Subscription:
    '''Tracks the status of a subscription.
    The ready property can be used to determine if a subscription is ready.
    The error property can be used to determine if a subscription encountered an error.
    '''
    
    def __init__(self, name):
        self._id = str(random.randint(0, 1000000))
        self._name = name
        
        self.ready = False
        self.error = None
        
        self._ready_event = asyncio.Event()

    def __ready__(self):
        self._ready_event.set()
    
    def __error__(self, error):
        self.error = error
        self._ready_event.set()
        
    async def wait(self):
        '''This coroutine waits for the subscription to become ready.
        If the server responds with an error then a ddp_asyncio.SubscriptionError will be raised.
        '''
        
        await self._ready_event.wait()
        if self.error: raise SubscriptionError(self.error.get('message', self.error))



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='ddp_asyncio',
    version='0.1.0',
    description='Asynchronous DDP library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/hunternet93/ddp_asyncio',
    download_url='https://github.com/hunternet93/ddp_asyncio/archive/ddp_asyncio-0.1.0.tar.gz',

    author='Isaac Smith',
    author_email='isaac@isrv.pw',

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',

        'License :: OSI Approved :: MIT License',
        
        'Operating System :: OS Independent',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        
        'Framework :: AsyncIO'
    ],

    keywords='ddp meteor',
    packages=find_packages(),
    install_requires=['websockets', 'ejson'],
)
